@@ -1,7 +1,12 @@
+#!/usr/bin/env python
+#
 # pg.py
+#
 # Written by D'Arcy J.M. Cain
 # Improved by Christoph Zwerschke
-# $Id: pg.py,v 1.33 2005/09/01 20:02:55 cito Exp $
+#
+# $Id: pg.py,v 1.48 2006/05/30 23:43:30 cito Exp $
+#
 
 """PyGreSQL classic interface.
 
@@ -10,6 +15,7 @@ It includes the _pg module and builds on it, providing the higher
 level wrapper class named DB with addtional functionality.
 This is known as the "classic" ("old style") PyGreSQL interface.
 For a DB-API 2 compliant interface use the newer pgdb module.
+
 """
 
 from _pg import *
@@ -36,7 +42,7 @@ def _quote(d, t):
 		return ("'f'", "'t'")[d]
 	if t in ('date', 'inet', 'cidr'):
 		if d == '': return 'NULL'
-	return "'%s'" % str(d).replace('\\','\\\\').replace('\'','\\\'')
+	return "'%s'" % str(d).replace("\\", "\\\\").replace("'", "''")
 
 def _is_quoted(s):
 	"""Check whether this string is a quoted identifier."""
@@ -114,6 +120,10 @@ class DB:
 		else:
 			raise InternalError, 'Connection is not valid'
 
+	# For convenience, define some module functions as static methods also:
+	escape_string, escape_bytea, unescape_bytea = map(staticmethod,
+		(escape_string, escape_bytea, unescape_bytea))
+
 	def _do_debug(self, s):
 		"""Print a debug message."""
 		if not self.debug: return
@@ -209,16 +219,26 @@ class DB:
 		then replace the __pkeys dictionary with it.
 
 		"""
-		# Get all the primary keys at once
+		# First see if the caller is supplying a dictionary
 		if isinstance(newpkey, DictType):
-			self.__pkeys = newpkey
-			return newpkey
+			# make sure that we have a namespace
+			self.__pkeys = {}
+			for x in newpkey.keys():
+				if x.find('.') == -1:
+					self.__pkeys['public.' + x] = newpkey[x]
+				else:
+					self.__pkeys[x] = newpkey[x]
+
+			return self.__pkeys
+
 		qcl = _join_parts(self._split_schema(cl)) # build qualified name
 		if newpkey:
 			self.__pkeys[qcl] = newpkey
 			return newpkey
+
+		# Get all the primary keys at once
 		if self.__pkeys == {} or not self.__pkeys.has_key(qcl):
-			# if not found, determine pkey again in case it was added after we started
+			# if not found, check again in case it was added after we started
 			for r in self.db.query("SELECT pg_namespace.nspname"
 				",pg_class.relname,pg_attribute.attname FROM pg_class"
 				" JOIN pg_namespace ON pg_namespace.oid=pg_class.relnamespace"
@@ -235,18 +255,35 @@ class DB:
 
 	def get_databases(self):
 		"""Get list of databases in the system."""
-		return [s for s, in
+		return [s[0] for s in
 			self.db.query('SELECT datname FROM pg_database').getresult()]
 
-	def get_tables(self):
-		"""Get list of tables in connected database."""
+	def get_relations(self, kinds = None):
+		"""Get list of relations in connected database of specified kinds.
+
+			If kinds is None or empty, all kinds of relations are returned.
+			Otherwise kinds can be a string or sequence of type letters
+			specifying which kind of relations you want to list.
+
+		"""
+		if kinds:
+			where = "pg_class.relkind IN (%s) AND" % \
+							','.join(["'%s'" % x for x in kinds])
+		else:
+			where = ''
+
 		return [_join_parts(s) for s in
-			self.db.query("SELECT pg_namespace.nspname"
-				",pg_class.relname FROM pg_class"
-				" JOIN pg_namespace ON pg_namespace.oid=pg_class.relnamespace"
-				" WHERE pg_class.relkind='r' AND"
-				" pg_class.relname!~'^Inv' AND "
-				" pg_class.relname!~'^pg_' ORDER BY 1,2").getresult()]
+			self.db.query(
+				"SELECT pg_namespace.nspname, pg_class.relname "
+				"FROM pg_class "
+				"JOIN pg_namespace ON pg_namespace.oid=pg_class.relnamespace "
+				"WHERE %s pg_class.relname !~ '^Inv' AND "
+					"pg_class.relname !~ '^pg_' "
+				"ORDER BY 1,2" % where).getresult()]
+
+	def get_tables(self):
+		"""Return list of tables in connected database."""
+		return self.get_relations('r')
 
 	def get_attnames(self, cl, newattnames = None):
 		"""Given the name of a table, digs out the set of attribute names.
@@ -268,7 +305,7 @@ class DB:
 		# May as well cache them:
 		if self.__attnames.has_key(qcl):
 			return self.__attnames[qcl]
-		if qcl not in self.get_tables():
+		if qcl not in self.get_relations('rv'):
 			raise ProgrammingError, 'Class %s does not exist' % qcl
 		t = {}
 		for att, typ in self.db.query("SELECT pg_attribute.attname"
@@ -277,40 +314,34 @@ class DB:
 			" JOIN pg_attribute ON pg_attribute.attrelid=pg_class.oid"
 			" JOIN pg_type ON pg_type.oid=pg_attribute.atttypid"
 			" WHERE pg_namespace.nspname='%s' AND pg_class.relname='%s'"
-			" AND pg_attribute.attnum>0 AND pg_attribute.attisdropped='f'"
+			" AND (pg_attribute.attnum>0 or pg_attribute.attname='oid')"
+			" AND pg_attribute.attisdropped='f'"
 				% cl).getresult():
-			if typ.startswith('interval'):
-				t[att] = 'date'
-			elif typ.startswith('int'):
-				t[att] = 'int'
+			if typ.startswith('bool'):
+				t[att] = 'bool'
 			elif typ.startswith('oid'):
 				t[att] = 'int'
-			elif typ.startswith('text'):
-				t[att] = 'text'
-			elif typ.startswith('char'):
-				t[att] = 'text'
-			elif typ.startswith('name'):
-				t[att] = 'text'
+			elif typ.startswith('float'):
+				t[att] = 'decimal'
 			elif typ.startswith('abstime'):
 				t[att] = 'date'
 			elif typ.startswith('date'):
 				t[att] = 'date'
+			elif typ.startswith('interval'):
+				t[att] = 'date'
+			elif typ.startswith('int'):
+				t[att] = 'int'
 			elif typ.startswith('timestamp'):
 				t[att] = 'date'
-			elif typ.startswith('bool'):
-				t[att] = 'bool'
-			elif typ.startswith('float'):
-				t[att] = 'decimal'
 			elif typ.startswith('money'):
 				t[att] = 'money'
 			else:
 				t[att] = 'text'
-		t['oid'] = 'int' # every table has this
 		self.__attnames[qcl] = t # cache it
 		return self.__attnames[qcl]
 
 	def get(self, cl, arg, keyname = None, view = 0):
-		"""Get a tuple from a database table.
+		"""Get a tuple from a database table or view.
 
 		This method is the basic mechanism to get a single row.  It assumes
 		that the key specifies a unique row.  If keyname is not specified
@@ -331,6 +362,13 @@ class DB:
 			keyname = self.pkey(qcl)
 		fnames = self.get_attnames(qcl)
 		if isinstance(arg, DictType):
+			# XXX this code is for backwards compatibility and will be
+			# XXX removed eventually
+			if not arg.has_key(foid):
+				ofoid = 'oid_' + self._split_schema(cl)[-1]
+				if arg.has_key(ofoid):
+					arg[foid] = arg[ofoid]
+
 			k = arg[keyname == 'oid' and foid or keyname]
 		else:
 			k = arg
@@ -403,6 +441,14 @@ class DB:
 		# otherwise use the primary key.  Fail if neither.
 		qcl = _join_parts(self._split_schema(cl)) # build qualified name
 		foid = 'oid(%s)' % qcl # build mangled oid
+
+		# XXX this code is for backwards compatibility and will be
+		# XXX removed eventually
+		if not a.has_key(foid):
+			ofoid = 'oid_' + self._split_schema(cl)[-1]
+			if a.has_key(ofoid):
+				a[foid] = a[ofoid]
+
 		if a.has_key(foid):
 			where = "oid=%s" % a[foid]
 		else:
@@ -433,10 +479,10 @@ class DB:
 		"""
 
 		This method clears all the attributes to values determined by the types.
-		Numeric types are set to 0, Booleans are set to 'f', dates are set
-		to 'now()' and everything else is set to the empty string.
-		If the array argument is present, it is used as the array and any entries
-		matching attribute names are cleared with everything else left unchanged.
+		Numeric types are set to 0, Booleans are set to 'f', and everything
+		else is set to the empty string.  If the array argument is present,
+		it is used as the array and any entries matching attribute names are
+		cleared with everything else left unchanged.
 
 		"""
 		# At some point we will need a way to get defaults from a table.
@@ -450,8 +496,6 @@ class DB:
 				a[k] = 0
 			elif t == 'bool':
 				a[k] = 'f'
-			elif t == 'date':
-				a[k] = 'now()'
 			else:
 				a[k] = ''
 		return a
@@ -468,6 +512,20 @@ class DB:
 		# isn't referenced somewhere (or else PostgreSQL will).
 		qcl = _join_parts(self._split_schema(cl)) # build qualified name
 		foid = 'oid(%s)' % qcl # build mangled oid
+
+		# XXX this code is for backwards compatibility and will be
+		# XXX removed eventually
+		if not a.has_key(foid):
+			ofoid = 'oid_' + self._split_schema(cl)[-1]
+			if a.has_key(ofoid):
+				a[foid] = a[ofoid]
+
 		q = 'DELETE FROM %s WHERE oid=%s' % (qcl, a[foid])
 		self._do_debug(q)
 		self.db.query(q)
+
+# if run as script, print some information
+if __name__ == '__main__':
+	print 'PyGreSQL version', version
+	print
+	print __doc__
